from __future__ import annotations

import os
import sys
import pathlib
import shutil
import stat

from typing import Final, Optional


def ansi(code):
    return f"\u001b[{code}m"

def _gen_appdata_folder() -> pathlib.Path:
    un = "faretek"
    match sys.platform:
        case "win32":
            return pathlib.Path(os.getenv('APPDATA')) / un
        case "linux":
            return pathlib.Path.home() / f".{un}"
        case plat:
            raise NotImplementedError(f"No 'appdata' folder implemented for {plat}")


APPDATA_FARETEK: Final[pathlib.Path] = _gen_appdata_folder()


APPDATA_FARETEK_INFLATE: Final[pathlib.Path] = APPDATA_FARETEK / "inflate"
APPDATA_FARETEK_COOKIES: Final[pathlib.Path] = APPDATA_FARETEK_INFLATE / "cookies.json"
APPDATA_FARETEK_PKGS: Final[pathlib.Path] = APPDATA_FARETEK_INFLATE / "pkgs"
APPDATA_FARETEK_ZIPAREA: Final[pathlib.Path] = APPDATA_FARETEK_INFLATE / "ziparea"

GITHUB_REPO: Final[str] = "https://github.com/FAReTek1/inflator"
AURA: Final[str] = "-9999 aura 💀"
ERROR_MSG: Final[str] = f"{ansi(31)}{AURA}{ansi(0)}\nFile an issue on github: {GITHUB_REPO}"


def rmtree(path, ignore_errors=False):
    # https://stackoverflow.com/questions/58878089/how-to-remove-git-repository-in-python-on-windows
    for root, dirs, files in os.walk(path):
        for d in dirs:
            os.chmod(os.path.join(root, d), stat.S_IRWXU)
        for file in files:
            os.chmod(os.path.join(root, file), stat.S_IRWXU)

    shutil.rmtree(path, ignore_errors=ignore_errors)



# store 'cookies' in appdata
import json

from typing import Any

import github

from inflator.util import APPDATA_FARETEK_COOKIES


class _Cookies:
    def __init__(self):
        (APPDATA_FARETEK_COOKIES / '..').resolve().mkdir(parents=True, exist_ok=True)
        if not APPDATA_FARETEK_COOKIES.exists():
            APPDATA_FARETEK_COOKIES.write_text("{}")

    @property
    def data(self) -> dict[str, str | int | None | bool | float | list | dict[str, Any]]:
        return json.load(APPDATA_FARETEK_COOKIES.open())

    @data.setter
    def data(self, data: dict[str, str | int | None | bool | float | list | dict[str, Any]]):
        json.dump(data, APPDATA_FARETEK_COOKIES.open("w"))

    def __setitem__(self, key: str, value: str | int | None | bool | float | list | dict[str, Any]):
        self.data |= {key: value}

    def __getitem__(self, key):
        return self.data[key]

    def __delitem__(self, key: str):
        self.data = {k: v for k, v in self.data.items() if k != key}

    def __contains__(self, item):
        return item in self.data

    def get(self, __key: str, __default=None):
        return self.data.get(__key, __default)


cookies = _Cookies()

gh = github.Github(cookies.get("auth-token"))



from __future__ import annotations

import logging

from inflator.util import APPDATA_FARETEK_PKGS
from inflator.package import Package


def install(raw: str, version: str = "*", *, upgrade: bool = False, ids: list[str] = None, editable: bool = False):
    logging.info(f"Raw install: {raw}")

    if ids is None:
        ids = []

    pkg = Package.from_raw(raw, version=version)
    logging.info(f"Parsed as {pkg}")
    pkg.install(ids=ids, editable=editable, upgrade=upgrade)



import shutil
import subprocess

from pathlib import Path
from typing import Optional

from slugify import slugify

from inflator.toml import toml as inflator_toml


def new(name: Optional[str] = None):
    if name is None:
        name = Path.cwd().name
        output_dir = Path.cwd()
    else:
        name = slugify(name)
        output_dir = Path(name)

    output_dir = output_dir.resolve()
    goboscript_output_dir = (output_dir / "..").resolve()

    print(f"Making {name!r}, dir={output_dir}")

    assert shutil.which("goboscript"), "You need to install goboscript: https://github.com/aspizu/goboscript/"
    if subprocess.call(["goboscript", "new", "-n", f"{name}"], cwd=str(goboscript_output_dir)) \
        == 0:
        inflator_toml(output_dir)



from __future__ import annotations

import fnmatch
import logging
import os
import pathlib
import hashlib
import pprint
import shutil
import tomllib

from dataclasses import dataclass, field
from typing import Optional, Self, Any
from zipfile import ZipFile
from io import BytesIO

import httpx

from furl import furl

from inflator.util import APPDATA_FARETEK_PKGS, APPDATA_FARETEK_ZIPAREA, rmtree
from inflator.parse import parse_iftoml, parse_gstoml
from inflator.cookies import gh
from inflator import gtp


@dataclass
class Package:
    username: Optional[str]
    reponame: str
    version: str

    raw: Optional[str] = None
    local_path: Optional[pathlib.Path] = None
    importname: Optional[str] = None
    is_local: Optional[bool] = None
    backpack_only: Optional[bool] = None

    deps: list[Package] = field(default_factory=list)

    @property
    def id(self):
        idstr = f"{self.username}\\{self.reponame}\\{self.version}"
        # hashing this to force use of this purely as an id, and not for string manipulation
        return hashlib.md5(idstr.encode()).hexdigest()

    @classmethod
    def from_raw(cls, raw: str, *, importname: Optional[str] = None, username: Optional[str] = None,
                 reponame: Optional[str] = None, version: str = '*',
                 _id: Optional[str] = None) -> Self:
        f = furl(raw)

        if f.host:
            assert f.host == 'github.com'  # Online packages must be from gh. no other website allowed.

            segments = f.path.normalize().segments
            if segments[-1] == '':
                segments = segments[:-1]

            assert len(segments) == 2
            _username, _reponame = segments
            local_path = None
            is_local = False
        else:
            local_path = pathlib.Path(raw).resolve()
            if local_path.exists():
                _username = None
                _reponame = local_path.parts[-1]
                is_local = True
            else:
                # try and fetch from the gtp
                print("Loading from gtp: ")
                data = gtp.load()
                if raw in data:
                    return cls.from_raw(data[raw]["url"], importname=importname, username=username, reponame=reponame, version=version, _id=_id)
                else:
                    raise FileNotFoundError(f"File {raw} not found")

        self = cls(
            username=_username,
            reponame=_reponame,
            version=version,
            importname=importname,
            local_path=local_path,
            raw=raw,
            is_local=is_local,
        )

        if self.id == _id and _id is not None:
            raise ValueError(f"Circular import of {self}")
        if username is not None:
            self.username = username
        if reponame is not None:
            self.reponame = reponame

        if self.is_local:
            self.resolve_toml_info()

        return self

    @property
    def name(self):
        return f"{self.reponame} {self.version} by {self.username}"

    @property
    def install_path(self):
        return APPDATA_FARETEK_PKGS / str(self.username) / self.reponame / self.version

    @property
    def zip_path(self):
        assert self.username is not None  # You can't have a GitHub package without a username
        return APPDATA_FARETEK_ZIPAREA / self.username / self.reponame / self.version

    def toml_path(self, name):
        return self.local_path / f"{name}.toml"

    def toml_file(self, name):
        return open(self.toml_path(name), "rb")

    def resolve_toml_info(self, _id: Optional[str] = None):
        assert self.local_path

        _id = self.id
        self.backpack_only = (not self.toml_path("inflator").exists()) and self.toml_path("goboscript").exists()

        logging.info("For self={}, iftoml_exists={}, gstoml_exists={}"
                     .format(self,
                             self.toml_path("inflator").exists(),
                             self.toml_path("goboscript").exists()))
        logging.info(f"So {self.backpack_only=}")

        if self.toml_path("inflator").exists():
            logging.info("Reading inflator.toml for name/version")

            data = parse_iftoml(tomllib.load(self.toml_file("inflator")), _id)

            if data.username:
                self.username = data.username
                logging.info(f"{self.username=}")

            if data.name:
                self.reponame = data.name
                logging.info(f"{self.reponame=}")

            if data.version:
                self.version = data.version
                logging.info(f"{self.version=}")

            self.deps += data.deps

        if self.toml_path("goboscript").exists():
            logging.info("Reading goboscript.toml for info")

            data = parse_gstoml(tomllib.load(self.toml_file("goboscript")), _id)

            logging.info(f"goboscript toml_data: {data}")

            self.deps += data.deps

        logging.info(f"Resolved {self.deps=}")

    def fetch_tag(self, pattern="*"):
        logging.info(f"Looking for tag for {self} with pattern {pattern}")
        assert not self.is_local

        logging.info(f"Fetching version name from github for {self}")

        tags = gh.get_repo(f"{self.username}/{self.reponame}").get_tags()

        logging.info(f"Collected tags: {pprint.pformat(tags)}")

        for tag in tags:
            name = tag.name
            if fnmatch.fnmatch(name, pattern):
                logging.info(f"Matched tag: {name}")
                return name

        if not tags:
            raise ValueError("No tags to match against.")
        else:
            raise ValueError("Matching tag could not be found, but alternatives available. Consider choosing {!r}"
                             .format(tags[-1].name))

    def fetch_data(self):
        logging.info(f"Trying to download {self} from gh")
        assert not self.is_local

        try:
            resp = httpx.get(
                f"https://api.github.com/repos/{self.username}/{self.reponame}/zipball/refs/tags/{self.version}",
                follow_redirects=True).raise_for_status()
        except httpx.HTTPError as e:
            e.add_note(f"Tag seems to be invalid. Maybe you meant {self.fetch_tag()!r}?")
            raise e

        logging.info(f"Downloaded {resp.content.__sizeof__()} bytes with status code {resp.status_code}")

        return resp.content

    @property
    def already_installed(self):
        pks = search_for_package(self.username, self.reponame, self.version)

        if pks:
            print("Found {} existing installation(s): {}"
                  .format(len(pks),
                          ''.join(f"\n- {pk.name}" for pk in pks)))
            return True

        return False

    def install(self, ids: Optional[list[str]] = None, editable: bool = False, upgrade: bool = False):
        if ids is None:
            ids = [self.id]
        elif self.id in ids:
            # not that you are allowed to depend on an old version of yourself
            raise RecursionError(f"Circular import of {self}")

        if self.is_local:
            if not upgrade:
                if self.already_installed:
                    return

            logging.info(f"Installing local package {self}")
            logging.info(f"Installing into {self.install_path}")

            if self.install_path.is_symlink():
                self.install_path.unlink()
            else:
                rmtree(self.install_path, ignore_errors=True)

            if editable:
                os.makedirs(pathlib.Path(*self.install_path.parts[:-1]), exist_ok=True)
                self.install_path.symlink_to(self.local_path)
            else:
                shutil.copytree(self.local_path, self.install_path, symlinks=True)

        else:
            logging.info(f"Installing gh package {self}")

            if not self.version:
                self.version = "*"
            self.version = self.fetch_tag(self.version)

            if not upgrade:
                if self.already_installed:
                    return

            zipball = self.fetch_data()

            shutil.rmtree(self.zip_path, ignore_errors=True)
            with ZipFile(BytesIO(zipball)) as archive:
                archive.extractall(self.zip_path)

            _, dirs, _ = next(self.zip_path.walk())
            extraction_path = self.zip_path / dirs[0]
            logging.info(f"Moving {extraction_path} to {self.install_path}")

            if self.install_path.is_symlink():
                self.install_path.unlink()
            else:
                rmtree(self.install_path, ignore_errors=True)
            shutil.move(extraction_path, self.install_path)
            rmtree(self.zip_path, ignore_errors=True)

            self.local_path = self.install_path
            self.resolve_toml_info()

        print(f"Collected {self.deps}")
        for dep in self.deps:
            # Don't pass in editable.
            dep.install(ids, upgrade=upgrade)

        print(f"Installed {self.name} into {self.install_path}")

    def resolve(self) -> Self:
        pkgs = search_for_package(self.username, self.reponame, self.version)
        if len(pkgs) > 1:
            logging.info(f"Multiple packages resolved for {self}\n\t{pkgs}")
        elif len(pkgs) == 0:
            logging.warning(f"Could not resolve {self}")
            return self

        pkg = pkgs[0]
        self.username = pkg.username
        self.reponame = pkg.reponame
        self.version = pkg.version
        self.backpack_only = pkg.backpack_only

        if not self.local_path:
            self.local_path = pkg.local_path

        logging.info(f"Resolved {self=}")

        return self

    @property
    def symlink_folder(self):
        return "backpack" if self.backpack_only else "inflator"


def search_for_package(usernames: Optional[list[str] | str] = None,
                       reponames: Optional[list[str] | str] = None,
                       versions: Optional[list[str] | str] = None,
                       globbed: bool = True):
    """
    Find all repos that fit the query. Uses globs unless specified otherwise
    :return: list[str] - list of string in format {username}\\{reponame}\\{version}
    """

    def handle_l(ls):
        # handle list so that it can work nicely. None -> [], single string -> [str]
        if isinstance(ls, str):
            ls = [ls]
        elif ls is None:
            ls = []

        return ls

    reponames = [r.lower() for r in handle_l(reponames)]
    versions = handle_l(versions)
    usernames = [u.lower() for u in handle_l(usernames)]

    logging.info(f"Searching for {reponames!r} {versions} by {usernames!r}")
    try:
        _, local_usernames, _ = next(APPDATA_FARETEK_PKGS.walk())

        results = []

        def match_l(pats, value):
            if globbed:
                return not pats or any(fnmatch.fnmatch(value, p) for p in pats)
            else:
                return not pats or value in pats

        for username in filter(lambda i: match_l(usernames, i.lower()), local_usernames):
            logging.info(f"\tSearching for {username=}")
            path1 = APPDATA_FARETEK_PKGS / username
            _, local_reponames, _ = next(path1.walk())

            for reponame in filter(lambda i: match_l(reponames, i.lower()), local_reponames):
                logging.info(f"\tSearching for {reponame=}")

                path2 = path1 / reponame
                _, local_versions, possible_symlinks = next(path2.walk())

                local_versions += [s for s in possible_symlinks if (path2 / s).is_symlink()]

                # Make newer versions come first.
                # Because v1.0.0 is before v0.0.0 in alphabetical order, we can just do a reverse string list sort
                local_versions.sort(reverse=True)

                for version in filter(lambda i: match_l(versions, i.lower()), local_versions):
                    logging.info(f"\tSearching for {version=}")

                    install_path = path2 / version
                    pkg = Package.from_raw(install_path, username=username, reponame=reponame, version=version)
                    results.append(pkg)

                    logging.info(f"\tGot {pkg}")
        return results
    except StopIteration as e:
        logging.info(f"Caught exception {e}")
        logging.info("Making appdata dirs because they don't seem to exist")

        os.makedirs(APPDATA_FARETEK_PKGS)
        return []



from __future__ import annotations

import json

from base64 import b64decode

from inflator.cookies import gh

_repo = None
def repo():
    global _repo

    if not _repo:
        _repo = gh.get_repo("inflated-goboscript/gtp")

    return _repo

def load() -> dict[str, dict[str, str]]:
    raw_data = repo().get_contents("gtp.json")
    assert raw_data is not None, "No data. Something went wrong with GitHub api. raise issue on gh"
    return json.loads(b64decode(raw_data.content).decode())



# If we cannot locate a inflator.toml file, try to resolve stuff from goboscript.toml
# Note that these packages will need to be imported into backpack/ instead of inflate/
from __future__ import annotations

import logging

from dataclasses import dataclass, field
from typing import Optional

from inflator import package


@dataclass
class IFToml:
    username: Optional[str] = None
    name: Optional[str] = None
    version: Optional[str] = None

    deps: list[package.Package] = field(default_factory=list)


def parse_gstoml(toml: dict, _id: Optional[str] = None):
    logging.info(f"Parsing gstoml {toml}")
    deps = []

    for name, src in toml.get("dependencies", {}).items():
        split = src.split("==")
        version = split[-1]
        url = '=='.join(split[:-1])

        if url == '':
            # we are probably dealing with an old goboscript package, using the '@' syntax
            split = src.split("@")
            version = split[-1]
            url = '@'.join(split[:-1])

        deps.append(package.Package.from_raw(url, version=version, importname=name))

    return IFToml(deps=deps)


def parse_iftoml(toml: dict, _id: Optional[str] = None):
    logging.info(f"Parsing iftoml {toml}")
    deps = []
    tdeps = toml.get("dependencies", {})

    for name, value in tdeps.items():
        if isinstance(value, str):
            raw, version = value, "*"
            username = None

        else:
            assert isinstance(value, list)

            if len(value) == 2:
                raw, version = value
                username = None
            elif len(value) == 3:
                raw, version, username = value
            else:
                raise ValueError(f"Unexpected {value=}")

        deps.append(
            package.Package.from_raw(raw, version=version, importname=name, _id=_id, username=username)
        )

    return IFToml(
        username=toml.get("username"),
        name=toml.get("name"),
        version=toml.get("version"),
        deps=deps
    )



from __future__ import annotations

import argparse
import pathlib
import pprint
import time
import tomllib
import logging

from datetime import datetime

from inflator import __version__
from inflator.install import install
from inflator.new import new as inflator_new
from inflator.parse import parse_gstoml, parse_iftoml
from inflator.package import search_for_package
from inflator.sync import sync
from inflator.util import ERROR_MSG, AURA
from inflator.toml import toml as inflator_toml
from inflator.cookies import cookies


def main():
    __file_dir__ = pathlib.Path(*pathlib.Path(__file__).parts[:-2])
    log_folder = __file_dir__ / "inflator-logs"

    log_folder.mkdir(exist_ok=True)

    logging.basicConfig(filename=log_folder / f"{time.time()}.log", level=logging.INFO)

    logging.info(f"init: {datetime.now()}")

    parser = argparse.ArgumentParser(
        prog="inflate",
        description="Manage libraries for use in goboscript.",
        epilog="When called with no args: Sync libraries in the current directory."
    )
    subparsers = parser.add_subparsers(dest="command")

    parser.add_argument("-i", "--input", action="store", help="Set input directory for syncing. Default is cwd")
    parser.add_argument("-V", "--version", action="store_true", dest="V", help="Get inflator version number")
    parser.add_argument("-L", "--log-folder", action="store_true", dest="L", help="Get log folder")

    install_parser = subparsers.add_parser("install", help="Install a package")
    install_parser.add_argument("parg", nargs="?", help="Package to install")
    install_parser.add_argument("-V", "--version", nargs="?", dest="install_version",
                                help="Version of package to install. Defaults to newest version")
    install_parser.add_argument("-e", "--editable", action="store_true", dest="install_editable",
                                help="For local packages. Whether to install the package as a symlink so that "
                                     "when you edit it, changes are directly made to your installation")
    install_parser.add_argument("-U", "--upgrade", action="store_true", dest="install_upgrade",
                                help="Whether to force overwrite packages "
                                     "instead of just installing ones that dont exist")
    install_parser.add_argument("-r", "--requirements", nargs="?", dest="install_requirements",
                                help="Path to inflator.toml/goboscript.toml file")

    find_parser = subparsers.add_parser("find", help="Locate a package with a name/version/creator. "
                                                     "Can also be used to list out installed pkgs. "
                                                     "It is globbed.")
    find_parser.add_argument("name", nargs="?", help="Name of package/repository")  # , dest="find_name")
    find_parser.add_argument("-V", "--version", nargs="?", dest="find_version",
                             help="Version of package")
    find_parser.add_argument("-U", "--username", nargs="?", dest="find_username",
                             help="Username of creator of package")

    parse_parser = subparsers.add_parser("parse", help="Parse gstoml or iftoml file")
    parse_parser.add_argument("name", nargs="?", help="Path to goboscript.toml or inflator.toml")  # , dest="find_name")

    toml_parser = subparsers.add_parser("toml", help="Add an inflator.toml file to cwd")

    new_parser = subparsers.add_parser("new", help="Create an (inflated) goboscript project")
    new_parser.add_argument("name", nargs="?", help="Name of package/repository")

    set_parser = subparsers.add_parser("set", help="Set config in cookies.json")
    set_parser.add_argument("key", help="Key of cookie")
    set_parser.add_argument("value", nargs='?', help="Value of cookie. Set empty to delete")

    # args, _ = parser.parse_known_args()
    args = parser.parse_args()

    match args.command:
        case "install":
            if args.install_requirements:
                with open(args.install_requirements, "rb") as f:
                    if f.name.endswith("goboscript.toml"):
                        deps = parse_gstoml(tomllib.load(f)).deps
                    elif f.name.endswith("inflator.toml"):
                        deps = parse_iftoml(tomllib.load(f)).deps
                    else:
                        raise ValueError(f"File {f.name!r} is not goboscript.toml or inflator.toml\n{ERROR_MSG}")

                for dep in deps:
                    dep.install(editable=args.install_editable, upgrade=args.install_upgrade)
            else:
                install(args.parg, args.install_version, upgrade=args.install_upgrade, editable=args.install_editable)

        case "find":
            pks = search_for_package(args.find_username, args.name, args.find_version)
            for pk in pks:
                print(pk.name)

        case "parse":
            with open(args.name, "rb") as f:
                if f.name.endswith("goboscript.toml"):
                    data = parse_gstoml(tomllib.load(f))
                else:
                    data = parse_iftoml(tomllib.load(f))
                pprint.pp(data)

        case "toml":
            inflator_toml()

        case "new":
            inflator_new(args.name)
        case "set":
            if args.value is None:
                print(f"Deleting {args.key!r}")
                del cookies[args.key]
            else:
                print(f"seting {args.key!r}={args.value!r}")
                cookies[args.key] = args.value

        case None:
            # no args
            if args.V:
                print(f"Inflate {__version__}")
            elif args.L:
                print(f"{log_folder=}")
            else:
                if args.input:
                    cwd = pathlib.Path(args.input)
                else:
                    cwd = pathlib.Path.cwd()

                sync(cwd)

        case _:
            print(f"Unknown command: {args.command!r}")



from __future__ import annotations

import logging
import pathlib
import shutil
import os

from inflator import package


def sync(path: pathlib.Path):
    logging.info(f"Collecting packages in {path}")
    pkg = package.Package.from_raw(str(path))

    def collect(_pkg: package.Package, *, toplevel=False):
        # Don't include the original directory, because that isn't a package - just something using them
        ret = [] if toplevel else [_pkg]
        for dep in _pkg.deps:
            dep.resolve()
            ret += collect(dep)

        return ret

    deps = collect(pkg, toplevel=True)

    print("Collected:{}"
          .format(''.join(f"\n- {dep.name}" for dep in deps) if deps else " nothing"))

    if not deps:
        print("Nothing to sync, so nothing to do")
        return

    (path / "backpack").mkdir(exist_ok=True)
    (path / "inflator").mkdir(exist_ok=True)

    shutil.rmtree(path / "backpack", ignore_errors=True)
    shutil.rmtree(path / "inflator", ignore_errors=True)

    for dep in deps:
        sympath = path / dep.symlink_folder / dep.importname

        logging.info(f"Symlinking {sympath} into {dep.install_path}")

        assert dep.install_path.exists()  # You haven't installed the dependency, or you may have spelt something wrong.

        os.makedirs(pathlib.Path(*sympath.parts[:-1]), exist_ok=True)
        sympath.unlink(missing_ok=True)
        sympath.symlink_to(dep.install_path, target_is_directory=True)



import setuptools
from inflator import __version__

setuptools.setup(
    name="inflator",
    version=__version__,
    packages=setuptools.find_packages(),

    entry_points={
        "console_scripts": [
            "inflate=inflator.__main__:main"
        ]
    },

    author="faretek1",
    description="Inflates gobos. A goboscript package manager.",
    long_description_content_type="text/markdown",
    long_description=open("README.md").read(),
    install_requires=open("requirements.txt").read(),
    keywords=["goboscript"],
    project_urls={
        "Source": "https://github.com/inflated-goboscript/inflator",
        "Documentation": "https://inflated-goboscript.github.io/inflator/",
        "Homepage": "https://inflated-goboscript.github.io/inflator/"
    }
)



from pathlib import Path

from slugify import slugify

from inflator.util import AURA


def toml(cwd: Path = None):
    if cwd is None:
        cwd = Path.cwd()

    fp = cwd / "inflator.toml"
    if fp.exists():
        print("Inflator.toml already exists!")
        return

    print(f"Creating {fp}")
    fp.write_text(f"""\
# inflator.toml syntax documentation: https://github.com/faretek1/inflator#inflator
name = "{slugify(fp.parts[-2])}"
version = "v0.0.0"
username = "if this is left blank then {AURA}"

[dependencies]
""", "utf-8")
